﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace TcpServerForSocketNonBlocking
{
    /// <summary>
    /// ノンブロッキングサーバ
    /// </summary>
    public partial class NBServerForm : Form
    {
        private const string START_MSG = "{0} [サーバ起動] IP：{1}、ポート：{2}で接続を開始しました。";
        private const string RECV_MSG = "{0} [データ受信] IP：{1}、ポート：{2}から接続されました。";
        private const string END_MSG = "{0} [サーバ停止] IP：{1}、ポート：{2}の接続を終了しました。";

        private BackgroundWorker listenWorker = null;
        private volatile bool isAliveServer = false;
        private static ManualResetEvent procWait = new ManualResetEvent(false);

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public NBServerForm()
        {
            InitializeComponent();
        }

        /// <summary>
        /// 開始ボタンクリックイベント
        /// </summary>
        /// <param name="sender">イベント発生オブジェクト</param>
        /// <param name="e">イベントデータ</param>
        private void startButton_Click(object sender, EventArgs e)
        {
            if (!this.isAliveServer)
            {
                if (this.listenWorker == null)
                {
                    // 接続先IP作成
                    IPAddress address = IPAddress.Parse(this.ipTextBox.Text);
                    IPEndPoint ipEndPoint = new IPEndPoint(address, Convert.ToInt32(this.portTextBox.Text));

                    this.listenWorker = new BackgroundWorker();
                    this.listenWorker.WorkerSupportsCancellation = true;
                    this.listenWorker.DoWork += new DoWorkEventHandler(listenWorker_DoWork);
                    this.listenWorker.RunWorkerAsync(ipEndPoint);
                }

                this.isAliveServer = true;
            }
        }

        /// <summary>
        /// 終了ボタンクリックイベント
        /// </summary>
        /// <param name="sender">イベント発生オブジェクト</param>
        /// <param name="e">イベントデータ</param>
        private void endButton_Click(object sender, EventArgs e)
        {
            this.TerminateServer();
        }

        /// <summary>
        /// 接続待ちバック処理イベント
        /// </summary>
        /// <param name="sender">イベント発生オブジェクト</param>
        /// <param name="e">イベントデータ</param>
        private void listenWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            BackgroundWorker innerWorker = sender as BackgroundWorker;

            TcpListener server = null;
            IPEndPoint ipEndPoint = e.Argument as IPEndPoint;
            try
            {
                server = new TcpListener(ipEndPoint);
                // 接続開始
                server.Start();
                this.ShowConnectStatus(START_MSG, ipEndPoint);

                while (!innerWorker.CancellationPending)
                {
                    // 接続待ちが存在する場合
                    if (server.Pending())
                    {
                        BackgroundWorker communicateWorker = new BackgroundWorker();
                        communicateWorker.WorkerSupportsCancellation = true;
                        communicateWorker.DoWork += new DoWorkEventHandler(communicateWorker_DoWork);
                        communicateWorker.RunWorkerAsync(server);
                        // 処理待機
                        procWait.WaitOne();
                    }
                    else
                    {
                        Thread.Sleep(100);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                e.Cancel = true;
                if (server != null)
                {
                    server.Stop();
                    server = null;
                    this.ShowConnectStatus(END_MSG, ipEndPoint);
                }
            }
        }

        /// <summary>
        /// 通信バック処理イベント
        /// </summary>
        /// <param name="sender">イベント発生オブジェクト</param>
        /// <param name="e">イベントデータ</param>
        private void communicateWorker_DoWork(object sender, DoWorkEventArgs e)
        {
            TcpListener server = e.Argument as TcpListener;
            TcpClient client = null;
            try
            {
                client = server.AcceptTcpClient();
                // 接続待ち処理再開状態にする
                procWait.Set();

                IPEndPoint remoteEndPoint = (IPEndPoint)client.Client.RemoteEndPoint;
                this.ShowConnectStatus(RECV_MSG, remoteEndPoint);

                using (NetworkStream stream = client.GetStream())
                {
                    // データ受信
                    List<byte> receiveAllBytes = new List<byte>();
                    byte[] receiveBytes = new byte[1024];
                    int readLength = 0;
                    while (stream.DataAvailable)
                    {
                        readLength = stream.Read(receiveBytes, 0, receiveBytes.Length);
                        if (readLength < receiveBytes.Length)
                        {
                            Array.Resize(ref receiveBytes, readLength);
                        }
                        receiveAllBytes.AddRange(receiveBytes);
                    }

                    string receiveData = Encoding.UTF8.GetString(receiveAllBytes.ToArray());
                    receiveData += ": Server接続成功";

                    // データ送信
                    byte[] sendBytes = Encoding.UTF8.GetBytes(receiveData);
                    stream.Write(sendBytes, 0, sendBytes.Length);
                    stream.Flush();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                if (client != null)
                {
                    client.Close();
                }
            }
        }

        /// <summary>
        /// 接続情報表示
        /// </summary>
        /// <param name="msg">メッセージ</param>
        /// <param name="endPoint">エンドポイント</param>
        private void ShowConnectStatus(string msg, IPEndPoint endPoint)
        {
            string showMsg = string.Format(msg, DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss"),
                endPoint.Address.ToString(), endPoint.Port);
            this.stsTextBox.BeginInvoke((Action<string>)this.ShowConnectStatus,
                new object[] { showMsg });
        }

        /// <summary>
        /// 接続情報表示
        /// </summary>
        /// <param name="text">表示テキスト</param>
        private void ShowConnectStatus(string text)
        {
            this.stsTextBox.Text += string.Format("{0}{1}", text, Environment.NewLine);
        }

        /// <summary>
        /// サーバ停止
        /// </summary>
        private void TerminateServer()
        {
            if (this.isAliveServer)
            {
                if (this.listenWorker != null)
                {
                    this.listenWorker.CancelAsync();
                    this.listenWorker = null;
                }

                this.isAliveServer = false;
            }
        }
    }
}
